package br.com.nubank.keyper;

import org.apache.commons.io.IOUtils;
import org.bouncycastle.asn1.ASN1Sequence;
import org.bouncycastle.asn1.oiw.OIWObjectIdentifiers;
import org.bouncycastle.asn1.x500.X500Name;
import org.bouncycastle.asn1.x509.*;
import org.bouncycastle.cert.X509ExtensionUtils;
import org.bouncycastle.cert.X509v3CertificateBuilder;
import org.bouncycastle.cert.jcajce.JcaX509CertificateConverter;
import org.bouncycastle.cert.jcajce.JcaX509ExtensionUtils;
import org.bouncycastle.crypto.util.PrivateKeyFactory;
import org.bouncycastle.operator.ContentSigner;
import org.bouncycastle.operator.DefaultDigestAlgorithmIdentifierFinder;
import org.bouncycastle.operator.DefaultSignatureAlgorithmIdentifierFinder;
import org.bouncycastle.operator.DigestCalculator;
import org.bouncycastle.operator.bc.BcDigestCalculatorProvider;
import org.bouncycastle.operator.bc.BcRSAContentSignerBuilder;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.math.BigInteger;
import java.nio.charset.StandardCharsets;
import java.security.KeyPair;
import java.security.KeyStore;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.cert.Certificate;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.util.Calendar;
import java.util.Date;
import java.util.TimeZone;
import java.util.UUID;

public final class Certificates {
    public final static char[] password = "uselesspw".toCharArray();

    public static PKCS12 from(byte[] pkcs12) {
        return new PKCS12(pkcs12);
    }

    public static PKCS12 from(InputStream pkcs12) {
        try {
            return from(IOUtils.toByteArray(pkcs12));
        } catch (IOException e) {
            throw new KeyperException(e);
        }
    }

    public static PKCS12 from(X509Certificate certificate, PrivateKey key) {
        try {
            ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
            KeyStore keyStore = KeyStore.getInstance("PKCS12");
            keyStore.load(null, password);
            keyStore.setCertificateEntry("alias", certificate);
            keyStore.setKeyEntry("alias", key, password, new Certificate[]{certificate});
            keyStore.store(outputStream, password);
            outputStream.flush();
            return from(outputStream.toByteArray());
        } catch (Exception e) {
            throw new KeyperException(e);
        }
    }

    public static PKCS12 generate(PKCS12 issuer, String subject) {
        KeyPair keyPair = Keys.generate();
        X509Certificate cert = genCertificate(issuer, subject, keyPair.getPublic());
        return from(cert, keyPair.getPrivate());
    }

    public static PKCS12 generate(PKCS12 issuer, String subject, BigInteger serialNumber, Date notBefore, Date notAfter) {
        KeyPair keyPair = Keys.generate();
        X509Certificate cert = genCertificate(issuer, subject, keyPair.getPublic(), serialNumber, notBefore, notAfter);
        return from(cert, keyPair.getPrivate());
    }

    public static KCertificate generate(PKCS12 issuer, String subject, PublicKey subjectPublicKey) {
        return new KCertificate(genCertificate(issuer, subject, subjectPublicKey));
    }

    public static KCertificate generate(PKCS12 issuer, String subject, PublicKey subjectPublicKey, BigInteger serialNumber, Date notBefore, Date notAfter) {
        return new KCertificate(genCertificate(issuer, subject, subjectPublicKey, serialNumber, notBefore, notAfter));
    }

    public static KCertificate from(String pemCert) {
        try {
            InputStream inputStream = new ByteArrayInputStream(pemCert.getBytes(StandardCharsets.UTF_8));
            CertificateFactory certFactory = CertificateFactory.getInstance("X.509");
            return new KCertificate((X509Certificate) certFactory.generateCertificate(inputStream));
        } catch (Exception e) {
            throw new KeyperException(e);
        }
    }

    private static X509Certificate genCertificate(PKCS12 issuer, String subject, PublicKey subjectPublicKey) {
        Date notBefore = new Date();
        Calendar c = Calendar.getInstance(TimeZone.getTimeZone("UTC"));
        c.setTime(notBefore);
        c.add(Calendar.DAY_OF_YEAR, 180);
        Date notAfter = c.getTime();
        BigInteger serialNumber = new BigInteger(UUID.randomUUID().toString().replaceAll("-", ""), 16);
        return genCertificate(issuer, subject, subjectPublicKey, serialNumber, notBefore, notAfter);
    }

    private static SubjectKeyIdentifier publicKeyToSubjectKeyIdentifier(PublicKey publicKey) {
        try {
            SubjectPublicKeyInfo pubKeyInfo = SubjectPublicKeyInfo.getInstance(publicKey.getEncoded());
            DigestCalculator digCalc = new BcDigestCalculatorProvider().get(new AlgorithmIdentifier(OIWObjectIdentifiers.idSHA1));
            X509ExtensionUtils utils = new X509ExtensionUtils(digCalc);
            return utils.createSubjectKeyIdentifier(pubKeyInfo);
        } catch (Exception e) {
            throw new KeyperException(e);
        }
    }

    private static AuthorityKeyIdentifier genAuthorityKeyIdentifier(X509Certificate cert) {
        try {
            DigestCalculator digCalc = new BcDigestCalculatorProvider().get(new AlgorithmIdentifier(OIWObjectIdentifiers.idSHA1));
            JcaX509ExtensionUtils utils = new JcaX509ExtensionUtils(digCalc);
            return utils.createAuthorityKeyIdentifier(cert.getPublicKey());
        } catch (Exception e) {
            throw new KeyperException(e);
        }
    }

    private static X509v3CertificateBuilder getCertBuilderWithExtensions(X509v3CertificateBuilder certBuilder, PublicKey publicKey, java.security.cert.Certificate certificate) {
        try {
            SubjectKeyIdentifier ski = publicKeyToSubjectKeyIdentifier(publicKey);
            certBuilder.addExtension(Extension.subjectKeyIdentifier, false, ski);

            X509Certificate cert = (X509Certificate) certificate;
            AuthorityKeyIdentifier aki = genAuthorityKeyIdentifier(cert);
            certBuilder.addExtension(Extension.authorityKeyIdentifier, false, aki);

            certBuilder.addExtension(Extension.basicConstraints, false, new BasicConstraints(false));

            KeyUsage ku = new KeyUsage(KeyUsage.digitalSignature | KeyUsage.dataEncipherment | KeyUsage.keyEncipherment | KeyUsage.keyAgreement);
            certBuilder.addExtension(Extension.keyUsage, true, ku);

            return certBuilder;
        } catch (Exception e) {
            throw new KeyperException(e);
        }
    }

    private static X509Certificate genCertificate(PKCS12 issuer, String subject, PublicKey subjectPublicKey, BigInteger serialNumber, Date notBefore, Date notAfter) {
        try {
            String alias = "alias";
            KeyStore issuerKeyStore = issuer.toKeyStore();

            X500Name issuerX500Name = new X500Name("C=BR, O=tcc-client-auth, CN=" + new KCertificate((X509Certificate) issuerKeyStore.getCertificate(alias)).getSubjectCN());
            PrivateKey issuerPrivateKey = (PrivateKey) issuerKeyStore.getKey(alias, password);

            X500Name subjectX500Name = new X500Name("C=BR, O=tcc-client-auth, CN=" + subject);

            byte[] pkEncoded = subjectPublicKey.getEncoded();
            SubjectPublicKeyInfo spki = new SubjectPublicKeyInfo(ASN1Sequence.getInstance(pkEncoded));

            X509v3CertificateBuilder certBuilder = new X509v3CertificateBuilder(issuerX500Name, serialNumber, notBefore, notAfter, subjectX500Name, spki);

            AlgorithmIdentifier sigAlgId = new DefaultSignatureAlgorithmIdentifierFinder().find("SHA256withRSA");
            AlgorithmIdentifier digAlgId = new DefaultDigestAlgorithmIdentifierFinder().find(sigAlgId);
            ContentSigner signer = new BcRSAContentSignerBuilder(sigAlgId, digAlgId).build(PrivateKeyFactory.createKey(issuerPrivateKey.getEncoded()));

            certBuilder = getCertBuilderWithExtensions(certBuilder, subjectPublicKey, issuerKeyStore.getCertificate(alias));
            JcaX509CertificateConverter converter = new JcaX509CertificateConverter();
            return converter.getCertificate(certBuilder.build(signer));
        } catch (Exception e) {
            throw new KeyperException(e);
        }
    }
}
