(ns service.service
  (:require
    [clojure.string :as s]
    [io.pedestal.http :as http]
    [io.pedestal.http.body-params :as body-params]
    [io.pedestal.http.route :as route]
    [ring.util.response :as ring-resp])
  (:import
    [br.com.nubank.keyper Certificates KCertificate Keys]
    [com.amazonaws.services.s3 AmazonS3Client]
    [com.amazonaws.services.s3.model GetObjectRequest]
    [org.apache.commons.io IOUtils]))

(defn get-object [bucket object-path]
  (let [client (AmazonS3Client.)
        request (GetObjectRequest. bucket object-path)
        response (.getObject client request)]
    response))

(defn get-object-byte-array [bucket object-path]
  (-> (get-object bucket object-path)
      .getObjectContent
      IOUtils/toByteArray))

(defn branch-from-s3 [branch-name]
  (get-object-byte-array "tcc-client-auth" (str "branches/" branch-name ".p12")))

(defn gen-certificate [request]
  (let [subject (-> request :path-params :id)
        pubkey (-> request :form-params :pubkey (s/replace "\"" ""))]
    {:status 200
     :body (-> (Certificates/generate (Certificates/from (branch-from-s3 "client")) subject (Keys/from pubkey)) .toPem)}))

(defn cert-info [cert]
  (str "Certificate presented:\nSubject: "
       (-> cert KCertificate. .getSubjectCN)
       "\nIssuer: "
       (-> cert KCertificate. .getIssuerCN)
       "\nSerial number: "
       (.getSerialNumber cert)))

(defn public [request]
  (let [cert (some-> request :servlet-request (.getAttribute "javax.servlet.request.X509Certificate") first)]
    {:status 200
     :body (if cert (cert-info cert) "No client certificate presented.")}))

(defn private [request]
  (let [cert (some-> request :servlet-request (.getAttribute "javax.servlet.request.X509Certificate") first)]
    (if cert
      {:status 200
       :body (cert-info cert)}
      {:status 401
       :body "ERROR: No certificate"})))

(defn home-page
  [request]
  (ring-resp/response "Hello World!"))

(def common-interceptors [(body-params/body-params) http/html-body])

(def routes #{["/" :get (conj common-interceptors `home-page)]
              ["/gen-certificate/:id" :post (conj [(body-params/body-params)] `gen-certificate)]
              ["/public" :get `public]
              ["/private" :get `private]})

(def service {:env :prod
              ::http/routes routes
              ::http/resource-path "/public"
              ::http/type :jetty
              ::http/port 8080
              ::http/container-options {:h2c? true
                                        :h2? false
                                        :ssl? false}})
